import torch
import numpy as np
from torch.utils.data.sampler import Sampler

class Hsampler(Sampler):
    """Homogeneously samples elements from each class.
    Ensuring each batch has balanced samples from each class.

    Notice: may discard some training samples if the original
    dataset is not homogenous across every class, therefore
    you might need more training epochs than vanilla networks 

    Arguments:
        data_labels (nparray,1D): labels of the dataset
        num_class (int): number of classes
    """

    def __init__(self,data_labels,num_class):
        self.data_labels = data_labels
        self.num_class = num_class
        idx_perclass = [0]*num_class
        num_perclass = [0]*num_class
        min_num_perclass = 1e10
        for class_idx in range(num_class):
            tmp = (data_labels==class_idx).nonzero()[0]
            idx_perclass[class_idx] = tmp
            num_perclass[class_idx] = len(tmp)
            if len(tmp)<min_num_perclass:
                min_num_perclass = len(tmp)
        self.idx_perclass = idx_perclass
        self.num_perclass = num_perclass
        self.min_num_perclass = min_num_perclass
        
    def __iter__(self):
        # needs to change for heterogeneous data
        new_idx_perclass=[0]*self.num_class
        for class_idx in range(self.num_class):
            new_idx = torch.randperm(self.num_perclass[class_idx])
            new_idx_perclass[class_idx] = (self.idx_perclass[class_idx])[new_idx][0:self.min_num_perclass]
        # transpose and flatten the matrix
        return iter(np.asarray(new_idx_perclass).transpose().reshape(-1))

    def __len__(self):
        # needs to change for heterogenous data
        #return(torch.LongTensor(self.num_perclass).sum())
        return(self.min_num_perclass*self.num_class)

