""" 
OLE loss and SVD output

Copyright (c)  Jose Lezama, 2017 
jlezama@fing.edu.uy
"""

import torch
from torch.autograd import Variable, Function
import scipy as sp
import scipy.linalg as linalg
import pdb

import numpy as np

# Inherit from Function
class OLESVD(Function):

    @staticmethod
    def forward(ctx, X, y):
        X = X.cpu().numpy()
        y = y.cpu().numpy()
        
        classes = np.unique(y)
        C = classes.size
        N, D = X.shape
        DELTA = 1.

        # gradients initialization
        Obj_c = 0
        dX_c = np.zeros((N,D))
        Obj_all =0
        dX_all = np.zeros((N,D))

        eigThd_compute = 1e-6 # threshold small eigenvalues for a better subgradient
        eigThd_output_ratio = 1e-1 # threshold small eigenvalues for output singular vectors V's
        #only eigvalues > 1e-3*max(eigvalues) will be counted as significant

        output_Vs = () # a tuple containing the truncated V's (truncated using eigThd_output)
        output_Ss = () # a tuple containing the singular values (for debuging only)        


        # compute objective and gradient for first term \sum ||X_c||_*
        for c in classes:
            A = X[y==c,:]
            
            # SVD
            U, S, V = sp.linalg.svd(A, full_matrices = False, lapack_driver='gesvd')
            V = V.T
            
            nuclear = np.sum(S)

            r_compute = np.sum(S<eigThd_compute)
            #r_compute = len((S<eigThd_compute).nonzero())
            r_output = np.sum(S<S[0]*eigThd_output_ratio)
            #r_output = len((S<S[0]*eigThd_output_ratio).nonzero())

            # L_c = max(DELTA, ||TY_c||_*)
            if nuclear>DELTA:
              Obj_c += nuclear # marker
            
              # discard small singular values
              uprod = U[:,0:U.shape[1]-r_compute].dot(V[:,0:V.shape[1]-r_compute].T)            
              dX_c[y==c,:]+=uprod
            else:
              Obj_c+= DELTA

            output_Vs += (torch.FloatTensor(V[:,0:V.shape[1]-r_output]).cuda(),)
            output_Ss += (torch.FloatTensor(S).cuda(),)
            
        # compute objective and gradient for secon term ||X||_*
        U, S, V = sp.linalg.svd(X, full_matrices = False, lapack_driver='gesvd')  # all classes
        V = V.T
        
        Obj_all = np.sum(S)

        r = np.sum(S<eigThd_compute)

        uprod = U[:,0:U.shape[1]-r].dot(V[:,0:V.shape[1]-r].T)

        dX_all = uprod

        # objective
        obj = (Obj_c  - Obj_all)/N

        # gradient
        dX = (dX_c  - dX_all)/N
        ctx.dX = torch.FloatTensor(dX)

        output_tuple = (torch.Tensor([obj]).cuda(),)+output_Vs+output_Ss # for debuging
        #output_tuple = (torch.Tensor([obj]).cuda(),)+output_Vs # for real application

        return output_tuple

    @staticmethod
    def backward(ctx, *grad_output_tuple):

        grad_obj = grad_output_tuple[0]
        return grad_obj*Variable(ctx.dX.cuda()), None



def prob_svd(X,Vs):
    '''
    Compute the projection of output to the singular vectors V's
    The inverse of squared distances are used as (unscaled) probabilities
    X          N x D
    Vs         touple of size K
    V(in Vs)   D x sime dimension r
    output     N x K
    '''
    THRESH = 1e-6
    N,D = X.shape
    X = X.t()
    inverse_dist = ()
    for V in Vs:
        proj = V.mm(V.t().mm(X))
        normed_proj = proj/(proj.norm(dim=0,keepdim=True)+THRESH)
        inverse_dist +=((X*normed_proj).sum(0).view(N,1),)
        '''
        coeff = V.t().mm(X)
        displacement = X-V.mm(coeff)
        #sq_distance = ((displacement**2).sum(0)).view(N,1)+THRESH
        #inverse_dist += (1/sq_distance,)
        sq_distance = ((displacement**2).sum(0)).view(N,1)
        inverse_dist +=(torch.exp(-sq_distance),)
        '''
    return torch.cat(inverse_dist,1)
